#include <math.h>
#include "tools.h"
#include "mpi.h"
#include "mpe_graphics.h"
#include "xtools/basex11.h"

/*@
    MPE_OpenGraphics - (collectively) open an X Windows display

    Input Parameters:
.   comm - Communicator of participating processes
.   display - Name of X window display.  If null, display will be taken from
    the DISPLAY variable on the process with rank 0 in comm.  If that is
    either undefined, or starts with w ":", then the value of display is
    `hostname`:0
.   x,y - position of the window.  If (-1,-1), then the user should be
    asked to position the window (this is a window manager issue).
.   w,h - width and height of the window, in pixels.
.   is_collective - true if the graphics operations are collective; this
    allows the MPE graphics operations to make fewer connections to the 
    display.  If false, then all processes in the communicator comm will 
    open the display; this could exceed the number of connections that your
    X window server allows.  Not yet implemented.

    Output Parameter:
.   handle - Graphics handle to be given to other MPE graphics routines.
@*/
int MPE_OpenGraphics( handle, comm, display, x, y, w, h, is_collective )
MPE_XGraph **handle;		/* handle to window                     */
MPI_Comm   comm;		/* communicator used for communication  */
char       *display;		/* display name, or NULL                */
int        x, y;		/* position of window, or -1,-1         */
int        w, h;		/* size of window                       */
int        is_collective;	/* whether this operation is collective */
{
Window     win;
MPE_XGraph *new;
int        myid, numprocs;

if (is_collective) {
    /* Not supported; just use individual connections */
    is_collective = 0;
    }

*handle            = 0;    /* In case of errors */

new		   = NEW(MPE_XGraph);    CHKPTR(new);
new->Cookie        = MPE_G_COOKIE;
new->comm	   = comm;
new->xwin          = NEW(XBWindow);      CHKPTR(new->xwin);
new->is_collective = is_collective;

MPI_Comm_size(comm,&numprocs);
MPI_Comm_rank(comm,&myid);

if (!display) {
    int str_len;
    if (myid == 0) {
	extern char *getenv();
	display = getenv( "DISPLAY" );
	if (!display || display[0] == ':') {
	    /* Replace display with hostname:0 */
	    display = (char *)malloc( 256 );
	    CHKPTRN( display );
	    SYGetHostName( display, 256-3 );
	    strcat( display, ":0" );
	    }
	str_len = strlen( display ) + 1;
	MPI_Bcast( &str_len, 1, MPI_INT, 0, comm );
	MPI_Bcast( display, str_len, MPI_CHAR, 0, comm );
	} 
    else {
	MPI_Bcast( &str_len, 1, MPI_INT, 0, comm );
	display = (char *) malloc( sizeof( char ) * str_len );
	MPI_Bcast( display, str_len, MPI_CHAR, 0, comm );
	}
    }

if (0 == myid) {
    XBQuickWindow( new->xwin, display, "MPE", x, y, w, h );
    /* ALWAYS send the local host */
    win = new->xwin->win;
    MPI_Bcast( &win, 1, MPI_INT, 0, comm );
    }
else {
    MPI_Bcast( &win, 1, MPI_INT, 0, comm );
    XBQuickWindowFromWindow( new->xwin, display, win );
    if (!new->xwin->win) {
	char myname[1024];
	SYGetHostName( myname, 1024 );
	fprintf( stderr, "Failed to connect from %s to %s\n", 
		 myname, display );
	}	 
    }
*handle = new;
return MPI_SUCCESS;
}

/*@
    MPE_DrawPoint - Draw a point on an X Windows display 

    Input Parameters:
.   handle - MPE graphics handle 
.   x,y - pixel position to draw.  Coordinates are upper-left origin (standard
    X11)
.   color - Color INDEX value.  See MPE_MakeColorArray.  By default, the colors
    MPE_WHITE, MPE_BLACK, MPE_RED, MPE_YELLOW, MPE_GREEN, MPE_CYAN, and 
    MPE_BLUE are defined.
@*/
int MPE_DrawPoint( handle, x, y, color )
MPE_XGraph *handle;
int        x, y;
MPE_Color  color;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
if (handle->is_collective) {
    }
else {
    XBSetPixVal( handle->xwin, handle->xwin->cmapping[color] );
    XDrawPoint( handle->xwin->disp, handle->xwin->win, 
	        handle->xwin->gc.set, x, y );
    }
return MPI_SUCCESS;
}

/*@
    MPE_DrawLine - Draw a line on an X11 display

    Input Parameters:
.   handle - MPE graphics handle 
.   x1,y1 - pixel position of one end of the line to draw.  Coordinates are 
            upper-left origin (standard X11)
.   x2,y2 - pixel position of the other end of the line to draw.  Coordinates 
            are upper-left origin (standard X11)
.   color - Color INDEX value.  See MPE_MakeColorArray.  By default, the colors
    MPE_WHITE, MPE_BLACK, MPE_RED, MPE_YELLOW, MPE_GREEN, MPE_CYAN, and 
    MPE_BLUE are defined.
@*/
int MPE_DrawLine( handle, x1, y1, x2, y2, color )
MPE_XGraph *handle;
int        x1, y1, x2, y2;
MPE_Color  color;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
if (handle->is_collective) {
    }
else {
    XBSetPixVal( handle->xwin, handle->xwin->cmapping[color] );
    XDrawLine( handle->xwin->disp, handle->xwin->win, 
	        handle->xwin->gc.set, x1, y1, x2, y2 );
    }
return MPI_SUCCESS;
}

/*@
    MPE_FillRectangle - Draw a filled rectangle on an X11 display 

    Input Parameters:
.   handle - MPE graphics handle 
.   x,y - pixel position of the upper left (low coordinate) corner of the 
            rectangle to draw.
.   w,h - width and height of the rectangle
.   color - Color INDEX value.  See MPE_MakeColorArray.  By default, the colors
    MPE_WHITE, MPE_BLACK, MPE_RED, MPE_YELLOW, MPE_GREEN, MPE_CYAN, and 
    MPE_BLUE are defined.

    Notes:
    This uses the X11 definition of width and height, so you may want to 
    add 1 to both of them.
@*/
int MPE_FillRectangle( handle, x, y, w, h, color )
MPE_XGraph *handle;
int        x, y, w, h;
MPE_Color  color;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
if (handle->is_collective) {
    }
else {
    XBSetPixVal( handle->xwin, handle->xwin->cmapping[color] );
    XFillRectangle( handle->xwin->disp, handle->xwin->win, 
	        handle->xwin->gc.set, x, y, w, h );
    }
return MPI_SUCCESS;
}

/*@
    MPE_Update - Update an X11 display

    Input Parameter:
.   handle - MPE graphics handle.

    Note:
    Only after an MPE_Update can you count on seeing the results of MPE 
    drawing routines.  This is caused by the buffering of graphics requests
    for improved performance.
@*/
int MPE_Update( handle )
MPE_XGraph *handle;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
if (handle->is_collective) {
    }
else {
    XFlush( handle->xwin->disp );
    }
return MPI_SUCCESS;
}

/* @
    MPE_CreateContour - 
@ */

/* @
    MPE_DrawContour - 
@ */

/*@
    MPE_CloseGraphics - Close a X11 graphics device

    Input Parameter:
.   handle - MPE graphics handle.
@*/
int MPE_CloseGraphics( handle )
MPE_XGraph *handle;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
FREE( handle );
return MPI_SUCCESS;
}

/*@
    MPE_MakeColorArray - Make an array of color indices

    Input Parameters:
.   handle - MPE graphics handle
.   nc     - Number of colors

    Output Parameter:
.   array - Array of color indices

    Notes:
    The new colors for a uniform distribution in hue space and replace the
    existing colors EXCEPT for MPE_WHITE and MPE_BLACK.
@*/
int MPE_MakeColorArray( handle, ncolors, array )
MPE_XGraph *handle;
int        ncolors;
MPE_Color  array[];
{
int i;
MPE_Color white;

if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
XBUniformHues( handle->xwin, ncolors + 2 );

/* XBUniformHues creates a spectrum like:
     BLACK red->yellow->blue->purple WHITE
   We want something like:
     WHITE BLACK red->yellow->blue->purple

   So I shifted things around a little.
*/

/* here's the original code: */
/*
for (i=0; i<ncolors; i++) 
    array[i] = handle->xwin->cmapping[i + 2];
*/

/* here's something that works: */
white = handle->xwin->cmapping[ncolors+1];
for (i=0; i<ncolors; i++) {
  array[i] = i+2;
  handle->xwin->cmapping[ncolors+1-i] = handle->xwin->cmapping[ncolors-i];
}
handle->xwin->cmapping[MPE_BLACK] = handle->xwin->cmapping[0];
handle->xwin->cmapping[MPE_WHITE] = white;

return MPI_SUCCESS;
}

/*@
  MPE_NumColors - Get the number of available colors

  Input Parameter:
. handle - MPE graphics handle

  Output Parameter:
. nc - Number of colors available on the display.
@*/
int MPE_NumColors( handle, nc )
MPE_XGraph *handle;
int        *nc;
{
if (handle->Cookie != MPE_G_COOKIE) {
    fprintf( stderr, "Handle argument is incorrect or corrupted\n" );
    return MPI_ERR_BAD_ARGS;
    }
*nc = handle->xwin->maxcolors;
return MPI_SUCCESS;
}

/*@
   MPE_DrawCircle - draw a circle

  Input Parameters:
. graph - MPE graphics handle
. centerx - horizontal center point of the circle
. centery - vertical center point of the circle
. radius - radius of the circle
. color - color of the circle
@*/
int MPE_DrawCircle( graph, centerx, centery, radius, color )
MPE_XGraph *graph;
int centerx, centery, radius;
MPE_Color color;
{
  int lasty, x, y, radius2;

  radius2 = radius*radius;
  lasty = 0;
  for (x=1-radius; x<=radius; x++) {
    y = sqrt( (double)(radius2-x*x) );
    MPE_DrawLine( graph, x-1+centerx,  lasty+centery, x+centerx,  y+centery,
		 color );
    MPE_DrawLine( graph, x-1+centerx, -lasty+centery, x+centerx, -y+centery,
		 color );
    lasty = y;
  }
  return MPI_SUCCESS;
}
