/*
 *  $Id: context_util.c,v 1.11 1994/04/01 23:34:44 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* context_util.c - Utilities used by the context chapter functions */

#include <mpi.h>


#define MPIR_MAX(a,b)  (((a)>(b))?(a):(b))

/* @

MPIR_Context_alloc - allocate some number of contexts over given communicator

@ */
int MPIR_Context_alloc ( comm, num_contexts, context )
MPI_Comm      comm;
int           num_contexts;
MPIR_CONTEXT *context;
{
  static MPIR_CONTEXT high_context = MPIR_FIRST_FREE_CONTEXT;
  int n, N, N2_prev, surfeit;
  int d, dst, src;
  MPIR_CONTEXT context_in;

  /* Intialize communicator size */
  (void) MPI_Comm_size ( comm, &N );
  
  /* If there's only one member, this is trivial */
  if ( N > 1 ) {

    /* Initialize variables */
    MPI_Status status;
    MPI_Comm comm_coll = comm->comm_coll;
    (void) MPI_Comm_rank ( comm, &n );
    N2_prev = comm->group->N2_prev;
    surfeit = N - N2_prev;

    /* Perform a combine operation to find maximum unused context */
    if ( n < N2_prev ) {
      if( n < surfeit ) {

	/* get the fanin letter from the upper "half" process: */
	dst = N2_prev + n;

        MPI_Recv( &context_in, 1, MPIR_CONTEXT_TYPE, dst, 0, comm_coll, &status );

	/* determine max context */
        high_context = MPIR_MAX( high_context, context_in );
      }

      /* combine on embedded N2_prev power-of-two processes */
      for (d = 1; d < N2_prev; d <<= 1) {
        dst = (n ^ d);

        /* Send and receive a context limit */
	MPI_Sendrecv( &high_context, 1, MPIR_CONTEXT_TYPE, dst, 0, 
		      &context_in, 1, MPIR_CONTEXT_TYPE, dst, 0, comm_coll, &status );

	/* determine max context */
        high_context = MPIR_MAX( high_context, context_in );
      }

      /* fanout data to nodes above N2_prev... */
      if ( n < surfeit ) {
	dst = N2_prev + n;
        MPI_Send( &high_context, 1, MPIR_CONTEXT_TYPE, dst, 0, comm_coll);
      }
    } 
    else {
      /* fanin data to power of 2 subset */
      src = n - N2_prev;
      MPI_Sendrecv ( &high_context, 1, MPIR_CONTEXT_TYPE, src, 0, 
		     &high_context, 1, MPIR_CONTEXT_TYPE, src, 0, comm_coll, &status);
    }
  } 
  
  /* Set the context and reset the highest context */
  (*context) = high_context;
  high_context += num_contexts;

  return(MPI_SUCCESS); 
}

/*+

MPIR_Context_dealloc - deallocate previously allocated contexts 

+*/
/*ARGSUSED*/
int MPIR_Context_dealloc ( comm, num, context )
MPI_Comm     comm;
int          num;
MPIR_CONTEXT context;
{
  /* This does nothing currently */
  return (MPI_SUCCESS);
}
