/*
 *  $Id: ic_start.c,v 1.6 1994/01/28 17:43:39 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>


/*@

MPI_Intercomm_start - create an intercommuncator from two intracommunicators

algorithm:
. 1) Allocate a "send_context"
. 2) Send context and lrank_to_grank list from local comm group 
     if I'm the local_leader.
@*/
int MPI_Intercomm_start ( local_comm, local_leader, peer_comm, 
                          remote_leader, tag, inter_request )
MPI_Comm local_comm;
int      local_leader;
MPI_Comm peer_comm;
int      remote_leader;
int      tag;
MPI_Intercomm_request *inter_request;
{
  MPIR_CONTEXT context;
  int local_size, local_rank, peer_size, peer_rank;
  int errno = MPI_SUCCESS;
  MPI_Intercomm_request i_req;

  /* Check for valid arguments to function */
  (void) MPI_Comm_size ( local_comm, &local_size );
  (void) MPI_Comm_rank ( local_comm, &local_rank );
  (void) MPI_Comm_size ( peer_comm,  &peer_size  );
  (void) MPI_Comm_rank ( peer_comm,  &peer_rank  );
  if (((local_comm    == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL)) ||
      ((peer_comm     == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL)) ||
      ((local_leader  >= local_size)  && (errno = MPI_ERR_UNKNOWN))   || 
      ((local_leader  <  0)           && (errno = MPI_ERR_UNKNOWN))   ||
      ((remote_leader >= peer_size)   && (errno = MPI_ERR_UNKNOWN))   || 
      ((remote_leader <  0)           && (errno = MPI_ERR_UNKNOWN))   ||
      ((peer_rank     == MPI_UNDEFINED_RANK) && (errno = MPI_ERR_RANK)))
    return (errno);

  /* Create MPI_Intercomm_request */
  i_req=(*inter_request)=(MPIR_Intercomm_request *)NEW(MPIR_Intercomm_request);
  i_req->tag = tag;
  i_req->local_comm = local_comm;
  i_req->local_size = local_size;
  i_req->local_leader = local_leader;
  i_req->peer_comm = peer_comm;
  i_req->remote_leader = remote_leader;

  /* Allocate a receive context for the inter-communicator */
  MPIR_Context_alloc ( local_comm, 1, &(i_req->recv_context) );

  /* If I'm the local leader, then exchange information */
  if (local_rank == local_leader) {
  
	/* Send the lrank_to_grank table of the local_comm and an allocated */
	/* context. Currently I use multiple messages to send this information. */
	/* Eventually, this will change(?) */
	MPI_Isend (&(i_req->local_size), 1, MPI_INT, remote_leader, tag, 
			   peer_comm, &(i_req->req[0]));
	MPI_Isend (&(i_req->recv_context), 1, MPIR_CONTEXT_TYPE, remote_leader, 
			   tag, peer_comm, &(i_req->req[1]));
	
	/* Post the receives for the information from the remote_leader */
	/* We don't post a receive for the remote group yet, because we */
	/* don't know how big it is yet. */
	MPI_Irecv (&(i_req->remote_size), MPI_INT, remote_leader, tag,
			   peer_comm, &(i_req->req[2]));
	MPI_Irecv (&(i_req->send_context), 1, MPIR_CONTEXT_TYPE, remote_leader,
			   tag, peer_comm, &(i_req->req[3]));
	
	/* We post this one last since this is the logical order in which it */
	/* should complete (sortof) */
	MPI_Isend (local_comm->group->lrank_to_grank, local_size, MPI_INT,
			   remote_leader, tag, peer_comm, (&i_req->req[4]));
  }

  return(errno);
}

