/*
 *  $Id: cart_shift.c,v 1.9 1994/04/13 13:27:36 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Cart_shift - Given a shift direction and amount, returns the
                 shifted source and destination ranks

Input Parameters:
. comm - communicator with cartesian structure (handle) 
. direction - coordinate dimension of shift (integer) 
. disp - displacement ($> 0$: upwards shift, $< 0$: downwards shift) (integer) 

Output Parameters:
. rank_source - rank of source process (integer) 
. rank_dest - rank of destination process (integer) 

@*/
int MPI_Cart_shift ( comm, direction, displ, source, dest )
MPI_Comm  comm;
int       direction;
int       displ;
int      *source;
int      *dest;
{
  int i, rank, size;
  int source_position, dest_position, save_position, periodic;
  int errno = MPI_SUCCESS;

  /* Check for valid arguments */
  if ( ((comm               == MPI_COMM_NULL) && (errno = MPI_ERR_COMM))     ||
       ((comm->topology.type!= MPI_CART)      && (errno = MPI_ERR_TOPOLOGY)) ||
       ((direction          <  0)             && (errno = MPI_ERR_ARG))      ||
       ((direction>=(comm->topology.tag.cart.ndims))&&(errno=MPI_ERR_ARG))   ||
       ((dest               == (int *)0)      && (errno = MPI_ERR_ARG))      ||
       ((source             == (int *)0)      && (errno = MPI_ERR_ARG))      )
    return MPIR_ERROR( comm, errno, "Error in MPI_CART_SHIFT" );

  /* Check the case for a 0 displacement */
  MPI_Comm_rank (comm, &rank);
  if (displ == 0) {
    (*source) = (*dest) = rank;
    return (errno);
  }

  /* Get ready for shifting */
  MPI_Comm_size (comm, &size);
  periodic = comm->topology.tag.cart.periods[direction];
  save_position = source_position = dest_position = 
    comm->topology.tag.cart.position[direction];
  
  /* Shift for the destination */
  dest_position += displ;
  if ( dest_position >= comm->topology.tag.cart.dims[direction] ) {
    if ( periodic )
      dest_position %= comm->topology.tag.cart.dims[direction];
    else
      dest_position = MPI_PROC_NULL;
  }
  else if ( dest_position < 0 ) {
    if ( periodic )
      dest_position += comm->topology.tag.cart.dims[direction];
    else
      dest_position = MPI_PROC_NULL;
  }
  comm->topology.tag.cart.position[direction] = dest_position;
  MPI_Cart_rank ( comm, comm->topology.tag.cart.position, dest );

  /* Shift for the source */
  source_position -= displ;
  if ( source_position >= comm->topology.tag.cart.dims[direction] ) {
    if ( periodic )
      source_position %= comm->topology.tag.cart.dims[direction];
    else
      source_position = MPI_PROC_NULL;
  }
  else if ( source_position < 0 ) {
    if ( periodic )
      source_position += comm->topology.tag.cart.dims[direction];
    else
      source_position = MPI_PROC_NULL;
  }
  comm->topology.tag.cart.position[direction] = source_position;
  MPI_Cart_rank ( comm, comm->topology.tag.cart.position, source );

  /* Restore my position */
  comm->topology.tag.cart.position[direction] = save_position;

  return (errno);
}


