/*
 *  $Id: cart_sub.c,v 1.9 1994/04/12 20:12:58 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <mpi.h>

/*@

MPI_Cart_sub - used to partition a communicator into subgroups which 
               form lower-dimensional cartesian subgrids

Input Parameters:
. comm - communicator with cartesian structure (handle) 
. remain_dims - the  i th entry of remain_dims specifies whether the   i th dimension is kept in the subgrid ( true ) or is dropped ( false ) (logical vector) 
Output Parameter:
. newcomm - communicator containing the subgrid that includes the calling process (handle) 

@*/
int MPI_Cart_sub ( comm, remain_dims, comm_new )
MPI_Comm  comm;
int      *remain_dims;
MPI_Comm *comm_new;
{
  int i, j, ndims;
  int num_remain_dims = 0;
  int remain_total = 1;
  int rank;
  int color = 0;
  int key = 0;
  int errno = MPI_SUCCESS;

  /* Check for valid arguments */
  if ( ((comm                == MPI_COMM_NULL)&&(errno = MPI_ERR_COMM)) ||
       ((comm->topology.type != MPI_CART)     &&(errno = MPI_ERR_TOPOLOGY))  ||
       ((remain_dims         == (int *)0)     &&(errno = MPI_ERR_ARG))      ||
       ((comm_new           == (MPI_Comm *)0) &&(errno = MPI_ERR_ARG))       )
    return MPIR_ERROR( comm, errno, "Error in MPI_CART_SUB" );
  
  /* Determine the remaining dimension total */
  ndims = comm->topology.tag.cart.ndims;
  for ( i=0; i<ndims; i++ )
    if ( remain_dims[i] ) {
      num_remain_dims++;
      remain_total *= comm->topology.tag.cart.dims[i];
	}

  /* Check for special case */
  if ( num_remain_dims == 0 ) {
    (*comm_new) = MPI_COMM_NULL;
    return (errno);
  }

  /* Split the old communicator */
  for ( i=0; i< ndims; i++ ) 
	if ( remain_dims[i] ) {
	  key = (key * comm->topology.tag.cart.dims[i]) +
		    comm->topology.tag.cart.position[i];
	}
	else {
	  color = (color * comm->topology.tag.cart.dims[i]) +
		      comm->topology.tag.cart.position[i];
	}
  MPI_Comm_split ( comm, color, key, comm_new );

  /* Store topology information in new communicator */
  if ( (*comm_new) != MPI_COMM_NULL ) {
    (*comm_new)->topology.type                    = MPI_CART;
    (*comm_new)->topology.tag.cart.nnodes         = remain_total;
    (*comm_new)->topology.tag.cart.ndims          = num_remain_dims;
    (*comm_new)->topology.tag.cart.dims           = 
      (int *)MPIR_MALLOC( sizeof(int) * 3 * num_remain_dims );
    (*comm_new)->topology.tag.cart.periods        = 
      (*comm_new)->topology.tag.cart.dims + num_remain_dims;
    (*comm_new)->topology.tag.cart.position       = 
      (*comm_new)->topology.tag.cart.periods + num_remain_dims;
    for ( i=j=0; i<ndims; i++ )
      if ( remain_dims[i] ) {
		(*comm_new)->topology.tag.cart.dims[j]    = 
		  comm->topology.tag.cart.dims[i];
		(*comm_new)->topology.tag.cart.periods[j] = 
		  comm->topology.tag.cart.periods[i];
		j++;
	  }
  
    /* Compute my position */
    MPI_Comm_rank ( (*comm_new), &rank );
    for ( i=0; i < num_remain_dims; i++ ) {
      remain_total = remain_total / (*comm_new)->topology.tag.cart.dims[i];
      (*comm_new)->topology.tag.cart.position[i]  = rank / remain_total;
      rank = rank % remain_total;
    }
  }
  return (errno);
}

